
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#define NPWM  400
#define PPWM  800

uint8_t f50,oen;
uint16_t pcount,uf;
uint32_t vpwr;
int sst;
int l[NPWM+1];
float pwr,ch2,ch0,pwm_trim;
uint16_t dbounce;
double Input, Output, Setpoint;
double errSum, lastErr,error,dErr;
double kp, ki, kd;
float xv[10],yv[10];

int vppwm,vppwm2;
long pcount_acc;
int pcount_delta;

void setup()
  {
  int i;
  float t,u;
 
  for(t=0.0,i=0; i <= NPWM; i++,t +=  2.0 * 3.14159/ (float) (NPWM+1))
    {
    u = 256.0 * sin(t);
    l[i] = (int)u;  
    }
  vppwm2 = 800;
  vppwm = 400;
  pcount_acc = 0;
  pcount_delta = 256;
  
  pinMode(9, OUTPUT);
  pinMode(10, OUTPUT);
  pinMode(8,INPUT_PULLUP);
  DDRD = 0xff;   // can NOT let arduino system alter any pins from D0 to D7 via calls to pinMode() etc.. It buggers d5 which I need to pull down low-Z upon boot
  noInterrupts();
  TCCR2A = 0; // I had some bits set which interfered with simple digital in PORTD. set to zero means no i/o pins driven by timer2.
  TCCR2B =   _BV(CS22)| _BV(CS21) | _BV(CS20);
  TIMSK2 |= (1 << TOIE2);
  TCCR1A = _BV(COM1A1) | _BV(COM1B0) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 10;
  ICR1 = vppwm;
  TIMSK1 |= (1 << TOIE1);
  
  cbi(PORTD,5);
  sst = 0;
  f50 = 0;
  pcount = 0;
  uf = 0;
  ch2 = ch0 = 0.0;
  pwr = 0.0;
  vpwr= 0;
  dbounce = 0;
  oen = 0;
    // set up PID 0.1 0.0 0.01 for biquad LP Fc = 0.02, np=4
    kp = 0.1;   
    ki = 0.0;
    kd = 0.01;
    errSum = 0.0;
 
  interrupts();             // enable all interrupts
  TIMSK0 = 0;                 // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;
  Serial.begin(115200);
  }


ISR(TIMER2_OVF_vect)
  { // 100Hz enable switch check, sets flag to run PID loop, ramp up etc.
  TCNT2=100;
  uf=1;
  check_switch();
  }

ISR(TIMER1_OVF_vect)        
 {
  // SPWM
  int c,d;
  uint16_t p2,p;


  c = vppwm + ((l[pcount] * vpwr) >> 8);   // scale sin wave by vpwr 32 bit integer calcs.

  p2 = pcount + NPWM/2;
  if (p2 >= NPWM)
    p2 -= NPWM;
    
  d = vppwm + ((l[p2] * vpwr) >> 8);   // scale sin wave by vpwr 32 bit integer calcs.
  
  OCR1A = c;
  OCR1B = vppwm2 - d -1;
  ICR1 = vppwm2;

  pcount_acc += pcount_delta;
  pcount = (pcount_acc >> 8);
  
  if(pcount >= NPWM)
    {
    pcount = 0;
    pcount_acc = 0;
    sbi(PORTD,7);  // give a sync pulse for DSO, alligned with start of 50Hz
    cbi(PORTD,7);
    }
}


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}

 
void loop()
{
  float s,fch0;
  // filter ch2,0 analog input and scale to 0 to 1
  // ch0 = output ACV, via transformer, 220VAC = 3.5V DC
  // ch2 = setpoint. About 4.0kHz sample rate, or 250us per 2ch sample and filter
  //ch2 = 0.99*ch2 + 0.01 * (float)analogRead(2)/1024.0;  // get AC output setpoint, LP filter

  pwm_trim = 0.99*pwm_trim + 0.01 * (float)analogRead(2)/1024.0; 

  
  ch2 = 0.57;
  fch0 = analogRead(0)/1024.0;  // get sample of AC output
  // this is a Bessel LP filter of some Fc.
  xv[0] = xv[1]; xv[1] = xv[2]; xv[2] = xv[3]; xv[3] = xv[4];
  xv[4] = fch0 / 2.259747797e+05;
  yv[0] = yv[1]; yv[1] = yv[2]; yv[2] = yv[3]; yv[3] = yv[4];
  yv[4] =   (xv[0] + xv[4]) + 4 * (xv[1] + xv[3]) + 6 * xv[2]
                     + ( -0.7428578687 * yv[0]) + (  3.1954036268 * yv[1])
                     + ( -5.1599230905 * yv[2]) + (  3.7073065280 * yv[3]);
  ch0 = yv[4];

  if(uf == 1  )    // this will execute at 100Hz, thanks to uf = 1, set within 100Hz ISR(TIMER2_OVF_vect)
      {                       // it runs at near zero volt crossing of AC output
      uf=0;
      
      if (oen == 1)        sst++;
      if (sst > 251)       sst = 251;  
      
      if (oen == 0)        sst--; 
      if (sst <= 0)
        {
          sst = 0;
          cbi(PORTD,5);
        }
      else
        sbi(PORTD,5);

      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {
        if (oen == 1) 
          { 
        if(ch0 < ch2)
          pwr += 1.0/250.0;
        else
          pwr -= 1.0/250.0;
          }
        else
          pwr -= 1.0 /250.0;
        }
        
      if (sst == 251)
        {
        Setpoint = ch2;     // PID control. set the setpoint, input get output, scale, convert to int
        Input = ch0;  
        do_pid();
        pwr = pwr + Output;
        }
        
      if (pwr > 1.0) pwr = 1.0;   // clamp . Do math in floating point to prevent any integer overflow
      if (pwr < 0.01) pwr = 0.01; // it's slower. We could do clamp after conversion to int but it's your mosfets...

      if (pwm_trim > 0.6)
        vppwm += 1;
      if (pwm_trim < 0.4)
        vppwm -= 1;
      if (vppwm < 200)
        vppwm = 200;


      vppwm2 = vppwm*2;
      s = vppwm * pwr;   
      vpwr = (int)s;
      s = 256.0 * (float)vppwm/400.0; 
      pcount_delta = (int)s;
      
      //Serial.println(OCR1A +OCR1B);
      //Serial.print(" ");
      Serial.println(vppwm);
      }
}

void check_switch()
{
  dbounce++;  
  if (dbounce > 100) dbounce = 100;
  if (digitalRead(8) == 0)     // momentary button pressed, pulling internal pullup low
    {
    if (dbounce > 80)
      {  
      if(oen == 0)
        {   // enable it. STAND BACK!
        vppwm = 400;
        vppwm2 = 800;
          
        oen = 1;
        }
      else
        {   // stop inverter
        oen = 0;
        }
     dbounce=0;
     }       
    }
}

